%% Copyright (c) Rodolà Emanuele - rodola@in.tum.de 
%  SHREC2016 - Partial Matching Of Deformable Shapes
%  http://www.dais.unive.it/~shrec2016/

function errors = calc_err_baryc(matches_in, matches_gt, M)

% M is the mesh in the canonical pose

if size(matches_in, 2)~=4 || size(matches_gt, 2)~=4
    error('The matches must be given in barycentric coordinates.');
end

valid_matches = matches_in(:,1)~=-1;
matches_in = matches_in(valid_matches,:);
matches_gt = matches_gt(valid_matches,:);

if ~all(abs(1-sum(matches_in(:,2:4),2))<1e-4)
    error('The matches must be given in barycentric coordinates with weights summing up to 1.');
end

if any(any(matches_in(:,2:4)<0))
    error('The barycentric coordinates must have non-negative weights.');
end

n = size(M.VERT,1);

% aux. structure for neighboring triangles
vertex_to_tri_list = cell(1,n);
for i=1:n
    vertex_to_tri_list{i} = find(M.TRIV(:,1)==i | M.TRIV(:,2)==i | M.TRIV(:,3)==i);
end

n_matches = size(matches_in,1);
errors = nan(n_matches,1);

geo_original = fastmarchmex('init', int32(M.TRIV-1), double(M.VERT(:,1)), double(M.VERT(:,2)), double(M.VERT(:,3)));

for i=1:n_matches
    
    if ~mod(i,n_matches/10)
%         fprintf('%.2f%%\n', 100*i/n_matches);
        fprintf('.');
    end    
%     fprintf('%d/%d\n', i, n_matches)
    
    from = 0;
    to = 0;
    
    t_gt_idx = matches_gt(i,1);
    t_in_idx = matches_in(i,1);
    
    t_gt = M.TRIV(t_gt_idx,:);
    t_in = M.TRIV(t_in_idx,:);
    
    % Check how and whether we will have to split the triangles
    
    on_vertex = find(matches_gt(i,2:4) > 0.99);
    if ~isempty(on_vertex)
        from = t_gt(on_vertex);
        p_gt = M.VERT(from,:);
    else
        [p_gt, edge_v1_gt, edge_v2_gt, edge_opposite_gt, split_in_triangle_gt] = ...
            check_split(M, matches_gt(i,:), t_gt);
    end
        
    on_vertex = find(matches_in(i,2:4) > 0.95);
    if ~isempty(on_vertex)
        to = t_in(on_vertex);
        p_in = M.VERT(to,:);
    else
        [p_in, edge_v1_in, edge_v2_in, edge_opposite_in, split_in_triangle_in] = ...
            check_split(M, matches_in(i,:), t_in);
    end
    
    % check if the input match is in the same triangle as the gt match
    if t_in_idx == t_gt_idx
        errors(i) = norm(p_gt-p_in);
        continue;
    end
    
    % check if we are matching original mesh vertices
    if from && to
        src = inf(n,1);
        src(from) = 0;
        d = fastmarchmex('march', geo_original, src);
        errors(i) = d(to);
        continue;
    end
    
    % Here the input match is in a different triangle than the ground truth
    
    if t_in_idx ~= t_gt_idx
        
        % approx. with euclidean distance if neighboring triangles
        if ~isempty(intersect(t_gt,t_in))
            %warning('Euclidean aproximation for adjacent triangles disabled');
            %errors(i) = norm(p_gt-p_in);
            %continue;
        end
        
        if (~from || ~to) % then at least one triangle must be split
            R = {};
            R.VERT = M.VERT;
            R.TRIV = M.TRIV;
        end
        
        % split triangles will be removed
        tris_to_remove = [];
        
        if ~from % the gt match is within a triangle or along an edge
            
            % add the new vertex
            R.VERT = [R.VERT ; p_gt];
            R.n = size(R.VERT,1);
            from = R.n;
            
            tris_to_remove = [tris_to_remove, t_gt_idx];
            
            if split_in_triangle_gt
                
                % split the current triangle into three triangles
                R.TRIV = [R.TRIV ; [R.n t_gt(1) t_gt(2)] ; [R.n t_gt(3) t_gt(1)] ; [R.n t_gt(2) t_gt(3)]];
                
            else
                
                % Split the current triangle into two triangles, making
                % sure the two new triangles have the right orientation.
                
                nn = cross(R.VERT(t_gt(2),:)-R.VERT(t_gt(1),:),R.VERT(t_gt(3),:)-R.VERT(t_gt(1),:));
                R.TRIV = [R.TRIV ; ...
                    add_split_tri(R, nn, [R.n edge_opposite_gt edge_v1_gt]) ; ...
                    add_split_tri(R, nn, [R.n edge_v2_gt edge_opposite_gt])];
                
                % detect the neighboring triangle
                tris = intersect(vertex_to_tri_list{edge_v1_gt} , vertex_to_tri_list{edge_v2_gt});
                
                if length(tris)>2 % non-manifold artifact
                    continue;
                end
                
                if length(tris)>1 % otherwise it's a boundary triangle,
                                  % there is nothing else to split
                    
                    neigh_tri_idx = tris(t_gt_idx~=tris);
                    neigh_tri = M.TRIV(neigh_tri_idx,:);
                    tris_to_remove = [tris_to_remove, neigh_tri_idx];
                
                    % split the neighboring triangle into two triangles
                    neigh_opposite = (neigh_tri~=edge_v1_gt & neigh_tri~=edge_v2_gt);
                
                    nn = cross(R.VERT(neigh_tri(2),:)-R.VERT(neigh_tri(1),:),R.VERT(neigh_tri(3),:)-R.VERT(neigh_tri(1),:));
                    R.TRIV = [R.TRIV ; ...
                        add_split_tri(R, nn, [R.n neigh_tri(neigh_opposite) edge_v1_gt]) ; ...
                        add_split_tri(R, nn, [R.n edge_v2_gt neigh_tri(neigh_opposite)])];
                    
                end
            end
        end
    
        if ~to % the input match is within a triangle or along an edge
            
            % add the new vertex
            R.VERT = [R.VERT ; p_in];
            R.n = size(R.VERT,1);
            to = R.n;
            
            tris_to_remove = [tris_to_remove, t_in_idx];
            
            if split_in_triangle_in
                
                % split the current triangle into three triangles
                R.TRIV = [R.TRIV ; [R.n t_in(1) t_in(2)] ; [R.n t_in(3) t_in(1)] ; [R.n t_in(2) t_in(3)]];
                
            else
                
                nn = cross(R.VERT(t_in(2),:)-R.VERT(t_in(1),:),R.VERT(t_in(3),:)-R.VERT(t_in(1),:));
                
                R.TRIV = [R.TRIV ; ...
                    add_split_tri(R, nn, [R.n edge_opposite_in edge_v1_in]) ; ...
                    add_split_tri(R, nn, [R.n edge_v2_in edge_opposite_in])];
                
                % detect the neighboring triangle
                tris = intersect(vertex_to_tri_list{edge_v1_in} , vertex_to_tri_list{edge_v2_in});

                if length(tris)>2
                    continue;
                end
                
                if length(tris)>1
                    
                    neigh_tri_idx = tris(t_in_idx~=tris);
                    neigh_tri = M.TRIV(neigh_tri_idx,:);
                    tris_to_remove = [tris_to_remove, neigh_tri_idx];
                    
                    % split the neighboring triangle into two triangles
                    neigh_opposite = (neigh_tri~=edge_v1_in & neigh_tri~=edge_v2_in);

                    nn = cross(R.VERT(neigh_tri(2),:)-R.VERT(neigh_tri(1),:),R.VERT(neigh_tri(3),:)-R.VERT(neigh_tri(1),:));
                    R.TRIV = [R.TRIV ; ...
                        add_split_tri(R, nn, [R.n neigh_tri(neigh_opposite) edge_v1_in]) ; ...
                        add_split_tri(R, nn, [R.n edge_v2_in neigh_tri(neigh_opposite)])];
                    
                end
                
            end
        end
        
        R.TRIV(tris_to_remove,:) = [];
        
        geo = fastmarchmex('init', int32(R.TRIV-1), double(R.VERT(:,1)), double(R.VERT(:,2)), double(R.VERT(:,3)));
        src = inf(R.n, 1);
        src(from) = 0;
        d = fastmarchmex('march', geo, src);
        errors(i) = d(to);
        fastmarchmex('deinit', geo);

    end
    
end
fprintf('\n');

errors = errors ./ sqrt( sum(calc_tri_areas(M)) );

fastmarchmex('deinit', geo_original);

end
